

typedef unsigned char   UBYTE;

// EquationList holds the RPN ROP equation. These are right out of the SDK,
// Programmer's Reference Vol. 3, Appendix A.
static char EquationList[][12]=
         {"0 ",         "DPSoon",       "DPSona",       "PSon",
          "SDPona",     "DPon",         "PDSxnon",      "PDSaon",

          "SDPnaa",     "PDSxon",       "DPna",         "PSDnaon",
          "SPna",       "PDSnaon",      "PDSonon",      "Pn",

          "PDSona",     "DSon",         "SDPxnon",      "SDPaon",
          "DPSxnon",    "DPSaon",       "PSDPSanaxx",   "SSPxDSxaxn",

          "SPxPDxa",    "SDPSanaxn",    "PDSPaox",      "SDPSxaxn",
          "PSDPaox",    "DSPDxaxn",     "PDSox",        "PDSoan",

          "DPSnaa",     "SDPxon",       "DSna",         "SPDnaon",
          "SPxDSxa",    "PDSPanaxn",    "SDPSaox",      "SDPSxnox",

          "DPSxa",      "PSDPSaoxxn",   "DPSana",       "SSPxPDxaxn",
          "SPDSoax",    "PSDnox",       "PSDPxox",      "PSDnoan",

          "PSna",       "SDPnaon",      "SDPSoox",      "Sn",
          "SPDSaox",    "SPDSxnox",     "SDPox",        "SDPoan",
          "PSDPoax",    "SPDnox",       "SPDSxox",      "SPDnoan",
          "PSx",        "SPDSonox",     "SPDSnaox",     "PSan",

          "PSDnaa",     "DPSxon",       "SDxPDxa",      "SPDSanaxn",
          "SDna",       "DPSnaon",      "DSPDaox",      "PSDPxaxn",
          "SDPxa",      "PDSPDaoxxn",   "DPSDoax",      "PDSnox",
          "SDPana",     "SSPxDSxoxn",   "PDSPxox",      "PDSnoan"
          ,
          "PDna",       "DSPnaon",      "DPSDaox",      "SPDSxaxn",
          "DPSonon",    "Dn",           "DPSox",        "DPSoan",
          "PDSPoax",    "DPSnox",       "DPx",          "DPSDonox",
          "DPSDxox",    "DPSnoan",      "DPSDnaox",     "DPan",

          "PDSxa",      "DSPDSaoxxn",   "DSPDoax",      "SDPnox",
          "SDPSoax",    "DSPnox",       "DSx",          "SDPSonox",
          "DSPDSonoxxn","PDSxxn",       "DPSax",        "PSDPSoaxxn",

          "SDPax",      "PDSPDoaxxn",   "SDPSnoax",     "PDSxnan",

          "PDSana",     "SSDxPDxaxn",   "SDPSxox",      "SDPnoan",
          "DSPDxox",    "DSPnoan",      "SDPSnaox",     "DSan",
          "PDSax",      "DSPDSoaxxn",   "DPSDnoax",     "SDPxnan",
          "SPDSnoax",   "DPSxnan",      "SPxDSxo",      "DPSaan"};

// Rops 128-255 are just like Rops 0-127, NOTed. The indexing is like so:
// IF (ROP>127) EquationList(ROP) = NOT(EquationList(255-ROP))

// EquationListLength is the length in characters of the RPN boolean equation
// that corresponds to the ROP.
static UBYTE EquationListLength[]={1,6,6,4,6,4,7,6,
                                   6,6,4,7,4,7,7,2,

                                   6,4,7,6,7,6,10,10,
                                   7,9,7,8,7,8,5,6,

                                   6, 6,4, 7,7,9,7,8,
                                   5,10,6,10,7,6,7,7,

                                   4,7,7,2,7,8,5,6,
                                   7,6,7,7,3,8,8,4,

                                   6, 6,7,9,4, 7,7,8,
                                   5,10,7,6,6,10,7,7,

                                   4,7,7,8,7,2,5,6,
                                   7,6,3,8,7,7,8,4,

                                    5,10,7, 6,7, 6,3,8,

                                   11, 6,5,10,5,10,8,7,

                                   6,10,7,7,7,7,8,4,
                                   5,10,8,7,8,7,7,6};


/*****************************************************************************/
/* DoRop                                                                     */
/*                                                                           */
/* DoRop is passed three bytes, one each from the current source,            */
/* destination, and pattern, and a ROP to use to combine the bytes. The      */
/* result of the ROP is returned.                                            */
/*                                                                           */
/* If this function is to be used with 24 bit values, it needs to be called  */
/* for each of the 3 color planes, then the results of the 3 calls are       */
/* combined.                                                                 */
/*****************************************************************************/
UBYTE DoRop(UBYTE src,    // The 8 bit value of the source
      UBYTE dest,         // The 8 bit value of the destination
      UBYTE pat,          // The 8 bit value of the brush
      UBYTE rop3)         // The ROP to use

{
    UBYTE Stack[7];       // Array used as a stack for calculating the RPN
                          // equation stored in EquationList that calculates
                          // the ROP.

    short StackPointer;  // Current position in the stack

    UBYTE a,b;            // Used to hold temporary results when executing
                          // an operator off the stack.

    UBYTE i;              // Loops over the RPN equation corresponding to the
                          // ROP

    UBYTE Rop3Index;      // Used to index into EquationList. EquationList
                          // has 128 ROPs in it, not the full 256. We take
                          // advantage of the fact that ROPs 128-256 are the
                          // same as ROPs 0-127, negated.

    short EquationLength; // Copy of equation length of current ROP
    UBYTE Equation[12];    // Copy of RPN equation for current ROP

    StackPointer = 0;   // Initialize stack pointer

    // Get length and boolean equation code from rop.
    // if ROP is 0-127, load it.
    // if ROP is 128-255, load 255-ROP.
    Rop3Index = (rop3 > 127) ? 255-rop3 : rop3;     // Get index for arrays

    EquationLength = EquationListLength[Rop3Index]; // Get ROP equation len.

    for (i=0; i< EquationLength; i++)               // Make a local copy of
    {                                               // the RPN ROP equation.
        Equation[i] = EquationList[Rop3Index][i];
    }

    if (rop3 > 127)  // If the ROP is > 127, we need to negate it.
    {
        if (Equation[EquationLength-1] == 'n') // If the equation happens
        {                                      // to end with a NOT operator,
            EquationLength--;                  // remove it by shortening the
        }                                      // equation length.

        else                                   // If the equation does not end
        {                                      // with a NOT operator, add one.
            Equation[EquationLength++] = 'n';
        }
    }

    for (i=0; i<EquationLength; i++)           // Loop over equation
    {
        if (Equation[i] == 'D')                // If Destination,
        {                                      // put current destination
            Stack[StackPointer++] = dest;      // value on the stack.

        }

        else if (Equation[i] == 'S')           // If Source,
        {                                      // put current source value
            Stack[StackPointer++] = src;       // on the stack
        }

        else if (Equation[i] == 'P')           // If pattern,
        {                                      // put current pattern value on
            Stack[StackPointer++] = pat;       // the stack.
        }

        else if (Equation[i] == '0')           // If zero, put zero on the
        {                                      // stack. This will only happen
            Stack[StackPointer++] = 0;         // with ROPs WHITENESS and
        }                                      // BLACKNESS.

        else if (Equation[i] == 'a')           // AND operator.
        {
            StackPointer--;                    // Pop stack twice into local
            a = Stack[StackPointer];           // variables, and push the
            b = Stack[StackPointer-1];         // smaller one back on the stack
            Stack[StackPointer-1] = (a < b) ? a : b;
        }

        else if (Equation[i] == 'n')           // NOT operator. Flip the bits
        {                                      // of the top of the stack.
            Stack[StackPointer-1] = (UBYTE)255 - Stack[StackPointer-1];
        }

        else if (Equation[i] == 'o')           // OR operator.
        {
            StackPointer--;                    // Pop stack twice into local
            a = Stack[StackPointer];           // variables, and push the
            b = Stack[StackPointer-1];         // larger one back on the stack
            Stack[StackPointer-1] = (a < b) ? b : a;
        }

        else  if (Equation[i] == 'x')          // XOR operator.
        {
            StackPointer--;                    // Pop stack twice into local
            a = Stack[StackPointer];           // variables, and push the ABS
            b = Stack[StackPointer-1];         // of the difference back on the
                                               // stack.
            Stack[StackPointer-1] = (a < b) ? b-a : a-b;
        }

    } // next i

    return Stack[0];                           // Top of stack now holds the
}                                              // desired value; return it.


